import json
import os
import re
import shutil
import tempfile
import time
import traceback
from datetime import datetime
from urllib.parse import urlsplit

from ... import util
from ...util import Info
from ...util.html import HtmlRewriter, Markup, MarkupTag
from ..host import Host

HTML_FILE_FILTER = re.compile(r'^.+\.x?html$', re.I)


# @TODO: better way to sync with WebScrapBook browser extension
ANNOTATION_CSS = """\
[data-scrapbook-elem="linemarker"][title] {
  cursor: help;
}
[data-scrapbook-elem="sticky"] {
  display: block;
  overflow: auto;
}
[data-scrapbook-elem="sticky"].styled {
  position: absolute;
  z-index: 2147483647;
  opacity: .95;
  box-sizing: border-box;
  margin: 0;
  border: 1px solid #CCCCCC;
  border-top-width: 1.25em;
  border-radius: .25em;
  padding: .25em;
  min-width: 6em;
  min-height: 4em;
  background: #FAFFFA;
  box-shadow: .15em .15em .3em black;
  font: .875em/1.2 sans-serif;
  color: black;
  overflow-wrap: break-word;
  cursor: help;
}
[data-scrapbook-elem="sticky"].styled.relative {
  position: relative;
  margin: 16px auto;
}
[data-scrapbook-elem="sticky"].styled.plaintext {
  white-space: pre-wrap;
}
[data-scrapbook-elem="sticky"].dragging {
  opacity: .75;
  z-index: 2147483641;
}
"""

ANNOTATION_JS = """\
function () {
  var w = window, d = document, r = d.documentElement, e;
  d.addEventListener('click', function (E) {
    if (r.hasAttribute('data-scrapbook-toolbar-active')) { return; }
    if (!w.getSelection().isCollapsed) { return; }
    e = E.target;
    if (e.matches('[data-scrapbook-elem="linemarker"]')) {
      if (e.title) {
        if (!confirm(e.title)) {
          E.preventDefault();
          E.stopPropagation();
        }
      }
    } else if (e.matches('[data-scrapbook-elem="sticky"]')) {
      if (confirm('%EditorDeleteAnnotationConfirm%')) {
        e.parentNode.removeChild(e);
        E.preventDefault();
        E.stopPropagation();
      }
    }
  }, true);
}
"""

BASIC_LOADER_JS = """\
function () {
  var k1 = "data-scrapbook-shadowdom",
      k2 = "data-scrapbook-canvas",
      k3 = "data-scrapbook-input-indeterminate",
      k4 = "data-scrapbook-input-checked",
      k5 = "data-scrapbook-option-selected",
      k6 = "data-scrapbook-input-value",
      k7 = "data-scrapbook-textarea-value",
      fn = function (r) {
        var E = r.querySelectorAll ? r.querySelectorAll("*") : r.getElementsByTagName("*"), i = E.length, e, d, s;
        while (i--) {
          e = E[i];
          if ((d = e.getAttribute(k1)) !== null && !e.shadowRoot && e.attachShadow) {
            s = e.attachShadow({mode: 'open'});
            s.innerHTML = d;
            e.removeAttribute(k1);
          }
          if ((d = e.getAttribute(k2)) !== null) {
            (function () {
              var c = e, g = new Image();
              g.onload = function () { c.getContext('2d').drawImage(g, 0, 0); };
              g.src = d;
            })();
            e.removeAttribute(k2);
          }
          if ((d = e.getAttribute(k3)) !== null) {
            e.indeterminate = true;
            e.removeAttribute(k3);
          }
          if ((d = e.getAttribute(k4)) !== null) {
            e.checked = d === 'true';
            e.removeAttribute(k4);
          }
          if ((d = e.getAttribute(k5)) !== null) {
            e.selected = d === 'true';
            e.removeAttribute(k5);
          }
          if ((d = e.getAttribute(k6)) !== null) {
            e.value = d;
            e.removeAttribute(k6);
          }
          if ((d = e.getAttribute(k7)) !== null) {
            e.value = d;
            e.removeAttribute(k7);
          }
          if (e.shadowRoot) {
            fn(e.shadowRoot);
          }
        }
      };
  fn(document);
}
"""


class Converter:
    def __init__(self, input, output, book_ids=None, *,
                 convert_legacy=False,
                 convert_v1=False,
                 use_native_tags=False,
                 ):
        self.input = input
        self.output = output
        self.book_ids = book_ids
        self.convert_legacy = convert_legacy
        self.convert_v1 = convert_v1
        self.use_native_tags = use_native_tags

    def run(self):
        if self.input != self.output:
            yield Info('info', 'Copying files...')
            self._copy_files()

        yield Info('info', 'Applying migration...')
        host = Host(self.output)

        # handle all books if none specified
        for book_id in self.book_ids or host.books:
            try:
                book = host.books[book_id]
            except KeyError:
                # skip invalid book ID
                yield Info('warn', f'Skipped invalid book "{book_id}".')
                continue

            yield Info('info', f'Handling book "{book_id}"...')
            book.load_meta_files()
            book.load_toc_files()

            if self.convert_legacy:
                yield Info('info', 'Migrating data files from legacy ScrapBook...')
                converter = ConvertDataFilesLegacy(book, use_native_tags=self.use_native_tags)
                yield from converter.run()

            if self.convert_v1:
                yield Info('info', 'Migrating to WebScrapBook 1.*...')
                converter = ConvertDataFilesV1(book, use_native_tags=self.use_native_tags)
                yield from converter.run()

    def _copy_files(self):
        with os.scandir(self.input) as dirs:
            for src in dirs:
                dst = os.path.join(self.output, src.name)
                try:
                    shutil.copytree(src, dst)
                except NotADirectoryError:
                    shutil.copy2(src, dst)


class ConvertDataFilesLegacy:
    """Convert data files with legacy data format.

    - Convert a web page with legacy annotations and chrome:// stylesheets or icons.
    - Convert a postit with legacy or other bad page wrapper.
    """
    def __init__(self, book, *, use_native_tags=False):
        self.book = book
        self.use_native_tags = use_native_tags

    def run(self):
        book = self.book
        for id, meta in book.meta.items():
            index = meta.get('index', '')
            if not index.endswith('/index.html'):
                continue

            type = meta.get('type', '')
            yield Info('debug', f'Converting data files for "{id}" (type="{type}")...')
            if type == 'postit':
                index_file = os.path.normpath(os.path.join(book.data_dir, index))
                yield Info('debug', f'Checking: {index_file}...')
                try:
                    content = book.load_note_file(index_file)
                except OSError as exc:
                    yield Info('error', f'Failed to convert "{index}" for "{id}": {exc.strerror}', exc=exc)
                else:
                    book.save_note_file(index_file, content)
            else:
                index_dir = os.path.normpath(os.path.dirname(os.path.join(book.data_dir, index)))
                for root, _dirs, files in os.walk(index_dir):
                    for file in files:
                        if HTML_FILE_FILTER.search(file):
                            file = os.path.join(root, file)
                            yield Info('debug', f'Checking: {file}...')
                            try:
                                conv = ConvertHtmlFileLegacy(
                                    file,
                                    use_native_tags=self.use_native_tags,
                                    host=self.book.host,
                                )
                                conv.run()
                            except Exception as exc:
                                traceback.print_exc()
                                yield Info('error', f'Failed to convert "{file}" for "{id}": {exc}', exc=exc)


class ConvertHtmlFileLegacy(HtmlRewriter):
    PRE_WRAP_REGEX = re.compile(r'\bwhite-space:\s*pre-wrap\b', re.I)

    LEGACY_CLASSES_MAP = {
        'linemarker-marked-line': 'linemarker',
        'scrapbook-inline': 'inline',
        'scrapbook-sticky': 'sticky',
        'scrapbook-sticky-header': 'sticky-header',
        'scrapbook-sticky-footer': 'sticky-footer',
        'scrapbook-block-comment': 'block-comment',
    }
    LEGACY_FREENOTE_STYLE_POSITION_REGEX = re.compile(r'position:\s*(\w+);', re.I)
    LEGACY_POS_STYLE_REGEX = re.compile(r'(?:left|top|width|height):\s*[\d.]+px;', re.I)
    LEGACY_COMBINE_CSS = """\
body {
    margin: 0px;
    background-color: #FFFFFF;
}
cite.scrapbook-header {
    clear: both;
    display: block;
    padding: 3px 6px;
    font-family: "MS UI Gothic","Tahoma","Verdana","Arial","Sans-Serif","Helvetica";
    font-style: normal;
    font-size: 12px;
    background-color: InfoBackground;
    border: 1px solid ThreeDShadow;
}
cite.scrapbook-header img {
    vertical-align: middle;
}
cite.scrapbook-header a {
    color: InfoText;
    text-decoration: none;
}
cite.scrapbook-header a[href]:hover {
    color: #3388FF;
}
cite.scrapbook-header a.marked { font-weight: bold; }
cite.scrapbook-header a.combine  { color: blue; }
cite.scrapbook-header a.bookmark { color: limegreen; }
cite.scrapbook-header a.notex { color: rgb(80,0,32); }
"""
    LEGACY_COMBINE_ICON_MAP = {
        'chrome://scrapbook/skin/treefolder.png': 'fclose.png',
        'chrome://scrapbook/skin/treenote.png': 'postit.png',
        'chrome://scrapbook/skin/treenotex.png': 'note.png',
        'chrome://scrapbook/skin/treeitem.png': 'item.png',
    }

    def __init__(self, file, *, use_native_tags=False, host=None):
        super().__init__(file)
        self.use_native_tags = use_native_tags
        self.host = host

    def run(self):
        self.require_annotation_loader = False
        self.combine_icons = {}
        self.map_id_markups = {}

        super().run()

    def rewrite(self, markups):
        for markup in markups:
            if markup.type == 'starttag':
                # record map of data-sb-id to markup
                id = markup.getattr('data-sb-id')
                self.map_id_markups.setdefault(id, []).append(markup)

                # check and record requirement of loader
                type = markup.getattr('data-scrapbook-elem')
                if type == 'sticky' or (type == 'linemarker' and markup.getattr('title') is not None):
                    self.require_annotation_loader = True

        markups, _ = self.convert(markups)

        # update annotation loader if there's a change
        if self.changed:
            markups = self._update_annotation_loaders(markups)

        return markups

    def convert(self, markups, start=0, endtag=None):
        rv = []
        i = start
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                # pass-through special context tags
                if markup.tag in ('template', 'svg', 'math'):
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023
                    for j in range(i, iend + 1):
                        rv.append(markups[j])
                    i = iend + 1
                    continue

                # handle legacy stylesheet
                if markup.tag == 'link' and markup.getattr('rel') == 'stylesheet':
                    href = markup.getattr('href')

                    # remove legacy annotation CSS
                    if href == 'chrome://scrapbook/skin/annotation.css':
                        i += 1
                        self.changed = True
                        continue

                    # convert legacy combine CSS
                    if href == 'chrome://scrapbook/skin/combine.css':
                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='starttag',
                            tag='style',
                            attrs=[
                                ('data-scrapbook-elem', 'custom-css'),
                            ],
                        ))
                        rv.append(Markup(
                            is_xhtml=self.is_xhtml,
                            type='data',
                            data=util.compress_code(self.LEGACY_COMBINE_CSS),
                            is_cdata=True,
                        ))
                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='endtag',
                            tag='style',
                        ))

                        i += 1
                        self.changed = True
                        continue

                # handle legacy combine icon
                if markup.tag == 'img':
                    src_changed = False
                    for j, attr_value in enumerate(markup.attrs):
                        attr, value = attr_value
                        if attr == 'src' and value in self.LEGACY_COMBINE_ICON_MAP:
                            new_src = self._get_legacy_combine_icon_url(value)
                            if new_src is not None:
                                markup.attrs[j] = (attr, new_src)
                                src_changed = True
                                break

                    if src_changed:
                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='starttag',
                            tag=markup.tag,
                            attrs=markup.attrs,
                        ))

                        i += 1
                        self.changed = True
                        continue

                # handle annotations
                type = self._get_legacy_scrapbook_object_type(markup)

                if type in ('linemarker', 'inline'):
                    tag = 'span' if self.use_native_tags else 'scrapbook-linemarker'
                    attrs = {
                        'data-scrapbook-id': None,
                        'data-scrapbook-elem': 'linemarker',
                        'class': [],
                    }

                    # id
                    id = markup.getattr('data-sb-id')
                    new_id = None

                    if id:
                        new_id = self._convert_legacy_scrapbook_elem_id(id)
                        try:
                            if markup == self.map_id_markups[id][0]:
                                attrs['class'].append('first')
                        except KeyError:
                            pass
                        try:
                            if markup == self.map_id_markups[id][-1]:
                                attrs['class'].append('last')
                        except KeyError:
                            pass

                    if new_id:
                        attrs['data-scrapbook-id'] = new_id

                    # style
                    css = markup.getattr('style')
                    if css:
                        attrs['style'] = css

                    # title
                    title = markup.getattr('title')
                    if title is not None:
                        attrs['title'] = title
                        self.require_annotation_loader = True

                    attrs['class'] = ' '.join(attrs['class'])
                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(
                        MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='starttag',
                            tag=tag,
                            attrs=attrs,
                        )
                    )

                    _rv, _i = self.convert(markups, i + 1, markup.endtag)
                    rv.extend(_rv)
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='endtag',
                        tag=tag,
                    ))

                    i = _i + 1
                    self.changed = True
                    continue

                elif type == 'freenote':
                    tag = 'div' if self.use_native_tags else 'scrapbook-sticky'
                    attrs = {
                        'data-scrapbook-elem': 'sticky',
                        'class': ['styled'],
                    }

                    # CSS
                    # @TODO: implement CSS parser for better error proof
                    css = markup.getattr('style')
                    if css:
                        m = self.LEGACY_FREENOTE_STYLE_POSITION_REGEX.search(css)
                        if m and m.group(1).lower() == 'static':
                            attrs['class'].append('relative')

                        css_new = ' '.join(m.group(0) for m in self.LEGACY_POS_STYLE_REGEX.finditer(css))
                        if css_new:
                            attrs['style'] = css_new

                    attrs['class'] = ' '.join(attrs['class'])
                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='starttag',
                        tag=tag,
                        attrs=attrs,
                    ))

                    _rv, _i = self.convert(markups, i + 1, markup.endtag)
                    rv.extend(_rv)
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='endtag',
                        tag=tag,
                    ))

                    i = _i + 1
                    self.changed = True
                    self.require_annotation_loader = True
                    continue

                elif type == 'sticky':
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023

                    tag = 'div' if self.use_native_tags else 'scrapbook-sticky'
                    attrs = {
                        'data-scrapbook-elem': 'sticky',
                        'class': ['styled', 'plaintext'],
                    }
                    if 'scrapbook-sticky-relative' in markup.classes:
                        attrs['class'].append('relative')

                    # CSS
                    # @TODO: implement CSS parser for better error proof
                    css = markup.getattr('style')
                    if css:
                        css_new = ' '.join(m.group(0) for m in self.LEGACY_POS_STYLE_REGEX.finditer(css))
                        if css_new:
                            attrs['style'] = css_new

                    # text content
                    textarea_i = self.find(markups, lambda x: x.tag == 'textarea', i + 1, markup.endtag)
                    if textarea_i is not None:
                        # unsaved sticky: take textarea content
                        textarea_iend = self.find(markups, lambda x: x == markups[textarea_i].endtag, textarea_i + 1, markup.endtag)  # noqa: B023
                        text = ''.join(str(d) for d in markups[textarea_i + 1:textarea_iend])
                    else:
                        last_child_i = i
                        for j in self.iterfind(markups, lambda x: x.type == 'endtag' and x != markup.endtag, i + 1, markup.endtag):  # noqa: B023
                            last_child_i = j
                        text = ''.join(str(d) for d in markups[last_child_i + 1:iend] if d.type == 'data')

                    attrs['class'] = ' '.join(attrs['class'])
                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='starttag',
                        tag=tag,
                        attrs=attrs,
                    ))
                    if text:
                        rv.append(Markup(
                            is_xhtml=self.is_xhtml,
                            type='data',
                            data=text,
                            convert_charrefs=False,
                        ))
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='endtag',
                        tag=tag,
                    ))

                    i = iend + 1
                    self.changed = True
                    self.require_annotation_loader = True
                    continue

                elif type == 'block-comment':
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023

                    tag = 'div' if self.use_native_tags else 'scrapbook-sticky'
                    attrs = {
                        'data-scrapbook-elem': 'sticky',
                        'class': ['plaintext', 'relative'],
                    }

                    # CSS
                    css = markup.getattr('style')
                    if css:
                        if not self.PRE_WRAP_REGEX.search(css):
                            css += ' white-space: pre-wrap;'
                    else:
                        css = 'white-space: pre-wrap;'
                    attrs['style'] = css

                    # text content
                    textarea_i = self.find(markups, lambda x: x.tag == 'textarea', i + 1, markup.endtag)
                    if textarea_i is not None:
                        # unsaved block-comment: take textarea content
                        textarea_iend = self.find(markups, lambda x: x == markups[textarea_i].endtag, textarea_i + 1, markup.endtag)  # noqa: B023
                        text = ''.join(str(d) for d in markups[textarea_i + 1:textarea_iend])
                    else:
                        text = ''.join(str(d) for d in markups[i + 1:iend] if d.type == 'data')

                    attrs['class'] = ' '.join(attrs['class'])
                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='starttag',
                        tag=tag,
                        attrs=attrs,
                    ))
                    if text:
                        rv.append(Markup(
                            is_xhtml=self.is_xhtml,
                            type='data',
                            data=text,
                            convert_charrefs=False,
                        ))
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='endtag',
                        tag=tag,
                    ))

                    i = iend + 1
                    self.changed = True
                    self.require_annotation_loader = True
                    continue

                else:
                    markup_changed = False
                    has_data_sb_obj = False
                    for j, attr_value in enumerate(markup.attrs):
                        attr, value = attr_value

                        # convert type attribute
                        if attr == 'data-sb-obj':
                            has_data_sb_obj = True
                            markup.attrs[j] = ('data-scrapbook-elem', type)
                            markup_changed = True

                        # convert id attribute
                        elif attr == 'data-sb-id':
                            markup.attrs[j] = ('data-scrapbook-id', self._convert_legacy_scrapbook_elem_id(value))
                            markup_changed = True

                        # convert data-sb-orig-<attr>
                        elif attr.startswith('data-sb-orig-'):
                            attr = f'data-scrapbook-orig-attr-{attr[13:]}'
                            markup.attrs[j] = (attr, value)
                            markup_changed = True

                    if type and not has_data_sb_obj:
                        markup.attrs.append(('data-scrapbook-elem', type))
                        markup_changed = True

                    if markup_changed:
                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='starttag',
                            tag=markup.tag,
                            attrs=markup.attrs,
                            is_self_end=markup.is_self_end,
                        ))

                        i += 1
                        self.changed = True
                        continue

            elif markup.type == 'endtag':
                if endtag is not None:
                    if markup == endtag:
                        break

            rv.append(markup)
            i += 1

        return rv, i

    def _get_legacy_scrapbook_object_type(self, markup):
        type = markup.getattr('data-sb-obj')
        if not type:
            for cls in markup.classes:
                if cls in self.LEGACY_CLASSES_MAP:
                    type = self.LEGACY_CLASSES_MAP[cls]
                    break
        return type

    def _get_legacy_combine_icon_url(self, src):
        if self.host is None or self.file is None:
            return None

        # attempt to load from cache
        try:
            return self.combine_icons[src]
        except KeyError:
            pass

        # save the icon under the root index dir
        root = os.path.dirname(self.file)
        filename, ext = os.path.splitext(os.path.basename(urlsplit(src).path))
        fsrc = self.host.get_static_file(self.LEGACY_COMBINE_ICON_MAP[src])
        fdst = os.path.join(root, f'{filename}{ext}')
        i = 1
        while os.path.lexists(fdst):
            fdst = os.path.join(root, f'{filename}-{i}{ext}')
            i += 1

        shutil.copyfile(fsrc, fdst)

        new_src = self.combine_icons[src] = util.get_relative_url(
            fdst,
            self.file,
            path_is_dir=False,
            start_is_dir=False,
        )

        return new_src

    def _convert_legacy_scrapbook_elem_id(self, id):
        """Convert legacy sb-obj-id (JavaScript timestamp) to WebScrapBook id.
        """
        try:
            dt = datetime.fromtimestamp(int(id) / 1000.0)
            return util.datetime_to_id(dt)
        except Exception:
            return id

    def _update_annotation_loaders(self, markups):
        # remove current loader
        rv = []
        i = 0
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                if markup.getattr('data-scrapbook-elem') in {'annotation-loader', 'annotation-css'}:
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023
                    i = iend + 1
                    continue

            rv.append(markup)
            i += 1

        # insert new loader
        if not self.require_annotation_loader:
            return rv

        markups = []

        markups.append(MarkupTag(
            is_xhtml=self.is_xhtml,
            type='starttag',
            tag='style',
            attrs=[
                ('data-scrapbook-elem', 'annotation-css'),
            ],
        ))
        markups.append(Markup(
            is_xhtml=self.is_xhtml,
            type='data',
            data=util.compress_code(ANNOTATION_CSS),
            is_cdata=True,
        ))
        markups.append(MarkupTag(
            is_xhtml=self.is_xhtml,
            type='endtag',
            tag='style',
        ))

        script = util.compress_code(ANNOTATION_JS)
        if self.host:
            script = util.format_string(script, self.host.get_i18n())
        script = f'({script})()'

        markups.append(MarkupTag(
            is_xhtml=self.is_xhtml,
            type='starttag',
            tag='script',
            attrs=[
                ('data-scrapbook-elem', 'annotation-loader'),
            ],
        ))
        markups.append(Markup(
            is_xhtml=self.is_xhtml,
            type='data',
            data=script,
            is_cdata=True,
        ))
        markups.append(MarkupTag(
            is_xhtml=self.is_xhtml,
            type='endtag',
            tag='script',
        ))

        pos = None
        for i in reversed(range(0, len(rv))):
            markup = rv[i]
            if markup.type == 'endtag':
                if markup.tag == 'body':
                    pos = i
                    break

                if markup.tag == 'html':
                    pos = i

        if pos is not None:
            rv = rv[:pos] + markups + rv[pos:]
        else:
            rv += markups

        return rv


class ConvertDataFilesV1:
    """Convert data files to latest WebScrapBook 1.*
    """
    def __init__(self, book, *, use_native_tags):
        self.book = book
        self.use_native_tags = use_native_tags

    def run(self):
        book = self.book
        for id, meta in book.meta.items():
            type = meta.get('type', '')
            if type == 'postit':
                continue

            index = meta.get('index', '')
            if not index:
                continue

            yield Info('debug', f'Converting data files for "{id}" (type="{type}")...')

            # folder
            if index.endswith('/index.html'):
                index_dir = os.path.normpath(os.path.dirname(os.path.join(book.data_dir, index)))
                for root, _dirs, files in os.walk(index_dir):
                    for file in files:
                        if HTML_FILE_FILTER.search(file):
                            file = os.path.join(root, file)
                            yield Info('debug', f'Checking: "{file}"...')
                            try:
                                conv = ConvertHtmlFileV1(file)
                                conv.run()
                            except Exception as exc:
                                traceback.print_exc()
                                yield Info('error', f'Failed to convert "{file}" for "{id}": {exc}', exc=exc)

            # htz/maff
            elif util.is_htz(index) or util.is_maff(index):
                index_file = os.path.normpath(os.path.join(book.data_dir, index))
                tempdir = tempfile.mkdtemp()
                tempzipdir = os.path.join(tempdir, 'zip')
                try:
                    util.zip_extract(index_file, tempzipdir)

                    changed = False
                    for root, _dirs, files in os.walk(tempzipdir):
                        for file in files:
                            if HTML_FILE_FILTER.search(file):
                                file = os.path.join(root, file)
                                subpath = file[len(tempzipdir) + 1:].replace('\\', '/')
                                yield Info('debug', f'Checking: "{subpath}" in "{index_file}"...')
                                try:
                                    conv = ConvertHtmlFileV1(file)
                                    conv.run()
                                    if conv.changed:
                                        changed = True
                                except Exception as exc:
                                    traceback.print_exc()
                                    yield Info('error', f'Failed to convert "{subpath}" in "{index_file}" for "{id}": {exc}', exc=exc)

                    # don't recompress (and change mtime) if no content changed
                    if changed:
                        util.zip_compress(index_file, tempzipdir, '')
                finally:
                    try:
                        shutil.rmtree(tempdir)
                    except OSError:
                        pass

            # single file
            elif util.is_html(index):
                file = os.path.normpath(os.path.join(book.data_dir, index))
                yield Info('debug', f'Checking: "{file}"...')
                try:
                    conv = ConvertHtmlFileV1(file)
                    conv.run()
                except Exception as exc:
                    traceback.print_exc()
                    yield Info('error', f'Failed to convert "{file}" for "{id}": {exc}', exc=exc)


class ConvertHtmlFileV1(HtmlRewriter):
    def run(self):
        self.require_basic_loader = False
        self.require_annotation_loader = False

        super().run()

    def rewrite(self, markups):
        markups = self.rewrite_doc(markups)

        # update loaders if there's a change
        if self.changed:
            markups = self._update_loaders(markups)

        return markups

    def rewrite_doc(self, markups):
        rv = []
        i = 0
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                # pass-through special context tags
                if markup.tag in ('template', 'svg', 'math'):
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023
                    for j in range(i, iend + 1):
                        rv.append(markups[j])
                    i = iend + 1
                    continue

                # check and record requirement of loader
                type = markup.getattr('data-scrapbook-elem')

                if type == 'basic-loader':
                    self.require_basic_loader = True

                if type in {'annotation-css', 'annotation-loader'}:
                    self.require_annotation_loader = True

                # handle old WebScrapBook loaders
                if markup.tag == 'script':
                    if markup.getattr('data-scrapbook-elem') in {
                        'canvas-loader',  # WebScrapBook < 0.69
                        'shadowroot-loader',  # WebScrapBook < 0.69
                    }:
                        iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023

                        i = iend + 1
                        self.changed = True
                        self.require_basic_loader = True
                        continue

                # handle old WebScrapBook attributes
                else:
                    for j, attr_value in enumerate(markup.attrs):
                        attr, value = attr_value
                        if attr == 'data-scrapbook-shadowroot':  # WebScrapBook < 0.115
                            value_new = self._convert_shadowroot_attribute(value)
                            markup.attrs[j] = ('data-scrapbook-shadowdom', value_new)
                            markup.src = None
                            self.changed = True
                            self.require_basic_loader = True
                            continue

            rv.append(markup)
            i += 1

        return rv

    def _update_loaders(self, markups):
        # remove current loader
        rv = []
        i = 0
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                if markup.getattr('data-scrapbook-elem') in {'basic-loader', 'annotation-loader', 'annotation-css'}:
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023
                    i = iend + 1
                    continue

            rv.append(markup)
            i += 1

        # insert new loader
        if self.require_basic_loader:
            markups = []

            script = util.compress_code(BASIC_LOADER_JS)
            script = f'({script})()'

            markups.append(MarkupTag(
                is_xhtml=self.is_xhtml,
                type='starttag',
                tag='script',
                attrs=[
                    ('data-scrapbook-elem', 'basic-loader'),
                ],
            ))
            markups.append(Markup(
                is_xhtml=self.is_xhtml,
                type='data',
                data=script,
                is_cdata=True,
            ))
            markups.append(MarkupTag(
                is_xhtml=self.is_xhtml,
                type='endtag',
                tag='script',
            ))

            pos = None
            for i in reversed(range(0, len(rv))):
                markup = rv[i]
                if markup.type == 'endtag':
                    if markup.tag == 'body':
                        pos = i
                        break

                    if markup.tag == 'html':
                        pos = i

            if pos is not None:
                rv = rv[:pos] + markups + rv[pos:]
            else:
                rv += markups

        if self.require_annotation_loader:
            markups = []

            markups.append(MarkupTag(
                is_xhtml=self.is_xhtml,
                type='starttag',
                tag='style',
                attrs=[
                    ('data-scrapbook-elem', 'annotation-css'),
                ],
            ))
            markups.append(Markup(
                is_xhtml=self.is_xhtml,
                type='data',
                data=util.compress_code(ANNOTATION_CSS),
                is_cdata=True,
            ))
            markups.append(MarkupTag(
                is_xhtml=self.is_xhtml,
                type='endtag',
                tag='style',
            ))

            script = util.compress_code(ANNOTATION_JS)
            if self.host:
                script = util.format_string(script, self.host.get_i18n())
            script = f'({script})()'

            markups.append(MarkupTag(
                is_xhtml=self.is_xhtml,
                type='starttag',
                tag='script',
                attrs=[
                    ('data-scrapbook-elem', 'annotation-loader'),
                ],
            ))
            markups.append(Markup(
                is_xhtml=self.is_xhtml,
                type='data',
                data=script,
                is_cdata=True,
            ))
            markups.append(MarkupTag(
                is_xhtml=self.is_xhtml,
                type='endtag',
                tag='script',
            ))

            pos = None
            for i in reversed(range(0, len(rv))):
                markup = rv[i]
                if markup.type == 'endtag':
                    if markup.tag == 'body':
                        pos = i
                        break

                    if markup.tag == 'html':
                        pos = i

            if pos is not None:
                rv = rv[:pos] + markups + rv[pos:]
            else:
                rv += markups

        return rv

    def _convert_shadowroot_attribute(self, data):
        markups = self.loads(json.loads(data)['data'])
        markups = self.rewrite_doc(markups)
        return ''.join(str(m) for m in markups if not m.hidden)


def run(input, output, book_ids=None, *,
        convert_legacy=False,
        convert_v1=False,
        use_native_tags=False,
        ):
    start = time.time()
    book_ids_text = ', '.join(f'"{id}"' for id in book_ids) if book_ids else 'all'
    yield Info('info', 'migrating:')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output) if output is not None else "(in-place)"}')
    yield Info('info', f'book(s): {book_ids_text}')
    yield Info('info', f'convert legacy: {convert_legacy}')
    yield Info('info', f'convert v1: {convert_v1}')
    yield Info('info', f'use native tags: {use_native_tags}')
    yield Info('info', '')

    if output is None:
        output = input

    try:
        conv = Converter(input, output, book_ids=book_ids,
                         convert_legacy=convert_legacy,
                         convert_v1=convert_v1,
                         use_native_tags=use_native_tags,
                         )
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import re
from urllib.parse import urljoin
from urllib.request import pathname2url

from . import util


class CssRewriter:
    """The base class that handles CSS rewriting for a reference path.
    """  # noqa: N815, F541
    pCm = f"""(?:/\\*[\\s\\S]*?(?:\\*/|$))"""  # comment  # noqa: N815, F541  # noqa: N815, F541
    pSp = f"""(?:[ \\t\\r\\n\\v\\f]*)"""  # space equivalents  # noqa: N815, F541
    pCmSp = f"""(?:(?:{pCm}|{pSp})*)"""  # comment or space  # noqa: N815, F541
    pCmSp2 = f"""(?:(?:{pCm}|{pSp})+)"""  # comment or space, at least one  # noqa: N815, F541
    pChar = f"""(?:\\\\.|[^\\\\"'])"""  # a non-quote char or an escaped char sequence  # noqa: N815, F541
    pStr = f"""(?:{pChar}*?)"""  # string  # noqa: N815, F541
    pSStr = f"""(?:{pCmSp}{pStr}{pCmSp})"""  # comment-or-space enclosed string  # noqa: N815, F541
    pDQStr = f"""(?:"[^\\\\"]*(?:\\\\.[^\\\\"]*)*")"""  # double quoted string  # noqa: N815, F541
    pSQStr = f"""(?:'[^\\\\']*(?:\\\\.[^\\\\']*)*')"""  # single quoted string  # noqa: N815, F541
    pES = f"""(?:(?:{pCm}|{pDQStr}|{pSQStr}|{pChar})*?)"""  # embeded string  # noqa: N815, F541
    pUrl = f"""(?:\\burl\\({pSp}(?:{pDQStr}|{pSQStr}|{pStr}){pSp}\\))"""  # URL  # noqa: N815, F541
    pUrl2 = f"""(\\burl\\({pSp})({pDQStr}|{pSQStr}|{pStr})({pSp}\\))"""  # URL; catch 3  # noqa: N815, F541
    pRImport = f"""(@import{pCmSp})({pUrl}|{pDQStr}|{pSQStr})"""  # @import; catch 2  # noqa: N815, F541
    pRFontFace = f"""(@font-face{pCmSp}{{{pES}}})"""  # @font-face; catch 1  # noqa: N815, F541
    pRNamespace = f"""(@namespace{pCmSp}(?:{pStr}{pCmSp2})?{pUrl})"""  # @namespace; catch 1  # noqa: N815, F541

    REGEX_REWRITE_CSS = re.compile(f"""{pCm}|{pRImport}|{pRFontFace}|{pRNamespace}|({pUrl})""", re.I)
    REGEX_PARSE_URL = re.compile(pUrl2, re.I)
    REGEX_UNESCAPE_CSS = re.compile(r"""\\(?:([0-9A-Fa-f]{1,6}) ?|(.))""")

    def __init__(self, file=None, *,
                 encoding=None,
                 ref_url=None, url_chain=set()):  # noqa: B006
        """Initialize the class and bind associated information.

        Args:
            file: path of the associated file.
            ref_url: overriding URL path of the reference path.
            encoding: the encoding for reading a file. None for autodetection
                (and self.encoding will be auto reset on reading).
        """
        self.file = file
        self.encoding = encoding
        self.ref_url = ref_url
        self.url_chain = url_chain.copy()

        if file:
            if not ref_url:
                self.ref_url = urljoin('file:///', pathname2url(file))

        if self.ref_url:
            self.url_chain.add(self.ref_url)

    def run(self, *args, **kwargs):
        """Common rewriting case when an associated file is provided.
        """
        if not self.file:
            raise RuntimeError('Associated file not set.')

        text = self.load(self.file)
        return self.rewrite(text, *args, **kwargs)

    def load(self, file):
        """Load a CSS file and return content text.

        May reset self.encoding.

        Args:
            file: str, path-like, or file-like bytes object
        Raises:
            OSError: failed to read the file
        """
        try:
            fh = open(file, 'rb')
        except TypeError:
            fh = file

        try:
            if not self.encoding:
                # Seek for the correct charset (encoding).
                # Use ISO-8859-1 if no encoding can be determined, so that
                # rewriting work independently with encoding.
                self.encoding = util.sniff_bom(fh) or 'ISO-8859-1'

            return fh.read().decode(self.encoding)
        finally:
            if fh != file:
                fh.close()

    def rewrite(self, text,
                rewrite_import_url=lambda url: url,
                rewrite_font_face_url=lambda url: url,
                rewrite_background_url=lambda url: url,
                ):
        def unescape_css(str):
            return self.REGEX_UNESCAPE_CSS.sub(unescape_css_sub, str)

        def unescape_css_sub(m):
            u, c = m.groups()
            if c:
                return c
            if u:
                return chr(int(u, 16))

        def parse_url(text, callback):
            def parse_url_sub(m):
                pre, url, post = m.groups()
                if url.startswith('"') and url.endswith('"'):
                    rewritten = callback(unescape_css(url[1:-1]))
                elif url.startswith("'") and url.endswith("'"):
                    rewritten = callback(unescape_css(url[1:-1]))
                else:
                    rewritten = callback(url.strip())
                return f'{pre}"{rewritten}"{post}'

            if not callback:
                return text

            return self.REGEX_PARSE_URL.sub(parse_url_sub, text)

        def rewrite_sub(m):
            im1, im2, ff, ns, u = m.groups()
            if im2:
                if im2.startswith('"') and im2.endswith('"'):
                    rewritten = rewrite_import_url(unescape_css(im2[1:-1]))
                elif im2.startswith("'") and im2.endswith("'"):
                    rewritten = rewrite_import_url(unescape_css(im2[1:-1]))
                else:
                    rewritten = parse_url(im2, rewrite_import_url)
                return f'{im1}"{rewritten}"'
            elif ff:
                return parse_url(ff, rewrite_font_face_url)
            elif ns:
                # do not rewrite @namespace rule
                return ns
            elif u:
                return parse_url(u, rewrite_background_url)
            return m.group(0)

        return self.REGEX_REWRITE_CSS.sub(rewrite_sub, text)



import os
import re
import shutil
import time
import traceback
from datetime import timedelta
from urllib.parse import quote
from urllib.request import pathname2url

from lxml import etree

from ... import util
from ...util import Info
from ...util.html import HtmlRewriter, Markup, MarkupTag
from ..book import Book
from ..host import Host

RDF = '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'
NS1 = '{http://amb.vis.ne.jp/mozilla/scrapbook-rdf#}'
NC = '{http://home.netscape.com/NC-rdf#}'

LEGACY_TYPE_MAP = {
    'postit': 'note',
    'note': 'notex',
}

HTML_FILE_FILTER = re.compile(r'^.+\.x?html$', re.I)
REGEX_LINEFEED = re.compile(r'\r\n?|\n')


class Converter:
    def __init__(self, input, output, book_id='', no_data_files=False):
        self.input = input
        self.output = output
        self.book_id = book_id
        self.no_data_files = no_data_files

        self.id_to_oid = {}
        self.oid_to_id = {}
        self.path_to_oid = {}
        self.icons_to_cache = {}

    def run(self):
        host = Host(self.input)

        try:
            book = host.books[self.book_id]
        except KeyError as exc:
            raise RuntimeError(f'book "{self.book_id}" does not exist') from exc

        book.load_meta_files()
        book.load_toc_files()

        yield Info('info', 'Creating ID mapping...')
        yield from self._generate_id_mapping(book)

        yield Info('info', 'Creating directory mapping...')
        yield from self._generate_directory_mapping(book)

        yield Info('info', 'Generating scrapbook.rdf...')
        yield from self._generate_rdf(book)

        yield Info('info', 'Copying data files...')
        yield from self._copy_data_files(book)

        yield Info('info', 'Copying icon files...')
        yield from self._copy_icon_files()

    def _generate_id_mapping(self, book):
        def get_legacy_id(id):
            ts = util.id_to_datetime(id)
            if ts:
                return util.datetime_to_id_legacy(ts)

            ts = util.id_to_datetime_legacy(id)
            if ts:
                return id

            return None

        def generate_unique_id(id):
            oid = get_legacy_id(id) or util.datetime_to_id_legacy()
            ts = util.id_to_datetime_legacy(oid)
            while oid in self.oid_to_id:
                ts += timedelta(seconds=1)
                oid = util.datetime_to_id_legacy(ts)
            return oid

        nonconvertable_ids = []

        # map convertable IDs
        for id in book.meta:
            yield Info('debug', f'Createding ID mapping: "{id}"')
            oid = get_legacy_id(id)
            if oid and oid not in self.oid_to_id:
                self.id_to_oid[id] = oid
                self.oid_to_id[oid] = id
                yield Info('debug', f'Created ID mapping: "{id}" => "{oid}"')
            else:
                nonconvertable_ids.append(id)

        # generate for non-convertable IDs
        for id in nonconvertable_ids:
            oid = generate_unique_id(id)
            self.id_to_oid[id] = oid
            self.oid_to_id[oid] = id
            yield Info('debug', f'Created ID mapping: "{id}" => "{oid}" (new)')

    def _generate_directory_mapping(self, book):
        for id, oid in self.id_to_oid.items():
            index = book.meta[id].get('index', '')
            if not index.endswith('/index.html'):
                continue

            dsrc = os.path.join(book.data_dir, os.path.dirname(index), '')
            self.path_to_oid[os.path.normcase(dsrc)] = oid
            yield Info('debug', f'Created directory mapping "{dsrc}" => "{oid}"')

    def _generate_rdf(self, book):
        def make_meta_node(id, meta):
            yield Info('debug', f'Generating meta node for "{id}"')

            oid = self.id_to_oid[id]
            type = meta.get('type', '')
            tagname = f'{NC}BookmarkSeparator' if type == 'separator' else f'{RDF}Description'

            node = etree.SubElement(root, tagname)
            node.attrib[f'{RDF}about'] = f'urn:scrapbook:item{oid}'
            node.attrib[f'{NS1}id'] = oid

            otype = LEGACY_TYPE_MAP.get(type, type)
            if meta.get('marked') and otype == '':
                node.attrib[f'{NS1}type'] = 'marked'
            else:
                node.attrib[f'{NS1}type'] = otype

            node.attrib[f'{NS1}title'] = meta.get('title', '')
            node.attrib[f'{NS1}chars'] = meta.get('charset', '')
            node.attrib[f'{NS1}source'] = meta.get('source', '')
            node.attrib[f'{NS1}icon'] = yield from self._handle_item_icon(book, id)
            node.attrib[f'{NS1}comment'] = meta.get('comment', '').replace('\n', ' __BR__ ')

            create = meta.get('create')
            if create:
                node.attrib[f'{NS1}create'] = util.datetime_to_id_legacy(util.id_to_datetime(create))

            modify = meta.get('modify')
            if modify:
                node.attrib[f'{NS1}modify'] = util.datetime_to_id_legacy(util.id_to_datetime(modify))

            locked = meta.get('locked')
            if locked is not None:
                node.attrib[f'{NS1}lock'] = 'true' if locked else ''

        def make_toc_node(id):
            yield Info('debug', f'Generating TOC node for "{id}"')

            node = etree.SubElement(root, f'{RDF}Seq')
            if id == 'root':
                node.attrib[f'{RDF}about'] = 'urn:scrapbook:root'
            else:
                oid = self.id_to_oid[id]
                node.attrib[f'{RDF}about'] = f'urn:scrapbook:item{oid}'

            for ref_id in book.toc.get(id, []):
                if ref_id in seen_in_toc:
                    yield Info('debug', f'Skipped adding "{ref_id}" under "{id}" (referenced)')
                    continue

                yield Info('debug', f'Adding "{ref_id}" under "{id}"')
                oref_id = self.id_to_oid[ref_id]
                child = etree.SubElement(node, f'{RDF}li')
                child.attrib[f'{RDF}resource'] = f'urn:scrapbook:item{oref_id}'
                seen_in_toc.add(ref_id)

                if ref_id in book.toc:
                    yield from make_toc_node(ref_id)

        root = etree.XML("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
</RDF:RDF>
""")

        for id, meta in book.meta.items():
            yield from make_meta_node(id, meta)

        seen_in_toc = {'root'}
        yield from make_toc_node('root')

        try:
            etree.indent(root)
        except AttributeError:
            # etree.indent is supported since lxml >= 4.5.0
            # Do simple line separation for first-level children for older
            # versions.
            for child in root:
                child.tail = '\n'

        tree = root.getroottree()
        file = os.path.join(self.output, 'scrapbook.rdf')
        os.makedirs(os.path.dirname(file), exist_ok=True)
        tree.write(file, encoding='UTF-8', xml_declaration=True, pretty_print=True)

    def _copy_data_files(self, book):
        for id, oid in self.id_to_oid.items():
            yield Info('debug', f'Copying data files for "{id}" => "{oid}"')
            type = book.meta[id].get('type', '')
            if type in Book.ITEM_TYPES_WITH_OPTIONAL_INDEX:
                yield Info('debug', f'Skipped copying data for "{id}": type is "{type}"')
                continue

            index = book.meta[id].get('index', '')
            if not index:
                yield Info('debug', f'Skipped copying data for "{id}": no index')
                continue

            try:
                if index.endswith('/index.html'):
                    fsrc = os.path.normpath(os.path.dirname(os.path.join(book.data_dir, index)))
                    fdst = os.path.join(self.output, 'data', oid)
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    shutil.copytree(fsrc, fdst)
                elif os.path.splitext(index)[1].lower() == '.html':
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid, 'index.html')
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    shutil.copy2(fsrc, fdst)
                elif util.is_htz(index):
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid)
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    util.zip_extract(fsrc, fdst)
                elif util.is_maff(index):
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid)
                    pages = util.get_maff_pages(fsrc)
                    if len(pages) == 0:
                        yield Info('error', f'Failed to copy data for "{id}": no page in MAFF archive')
                        continue
                    else:
                        if len(pages) > 1:
                            yield Info('warn', f'"{id}": only the first web page in MAFF archive can be copied')
                        page = pages[0]
                        topdir, _, _ = page.indexfilename.partition('/')

                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    util.zip_extract(fsrc, fdst, topdir)
                else:
                    basename = os.path.basename(index)
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid, basename)
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    shutil.copy2(fsrc, fdst)
                    with open(os.path.join(self.output, 'data', oid, 'index.html'), 'w', encoding='UTF-8') as fh:
                        fh.write('<html><head><meta charset="UTF-8">'
                                 f'<meta http-equiv="refresh" content="0;URL=./{quote(basename)}">'
                                 '</head><body></body></html>')
            except OSError as exc:
                yield Info('error', f'Failed to copy data for "{id}": {exc}', exc=exc)

            if type == 'postit':
                yield Info('debug', f'Converting data file for "{id}" (type={type})')
                file = os.path.join(self.output, 'data', oid, 'index.html')
                try:
                    content = book.load_note_file(file)
                except OSError as exc:
                    yield Info('error', f'Failed to convert "index.html" for "{id}": {exc.strerror}', exc=exc)
                else:
                    with open(file, 'w', encoding='UTF-8') as fh:
                        fh.write(f"""\
<html><head><meta http-equiv="Content-Type" content="text/html;Charset=UTF-8"></head><body><pre>
{content}
</pre></body></html>""")
            elif not self.no_data_files:
                yield Info('debug', f'Converting data files for "{id}" (type={type})')
                index_dir = os.path.join(self.output, 'data', oid)
                for root, _dirs, files in os.walk(index_dir):
                    for file in files:
                        if HTML_FILE_FILTER.search(file):
                            file = os.path.join(root, file)
                            yield Info('debug', f'Checking: {file}...')
                            try:
                                conv = ConvertHtmlFile(file)
                                conv.run()
                            except Exception as exc:
                                traceback.print_exc()
                                yield Info('error', f'Failed to convert "{file}" for "{id}": {exc}', exc=exc)

    def _handle_item_icon(self, book, id):
        yield Info('debug', f'Checking icon for "{id}"')
        icon = book.meta[id].get('icon', '')

        if not icon:
            # return moz_icon if defined
            moz_icon_url = book.meta[id].get('icon-moz')
            if moz_icon_url:
                yield Info('debug', f'Use moz-icon URL from property for "{id}": "{moz_icon_url}"')
                return moz_icon_url

            # generate moz-icon:// for files
            if book.meta[id].get('type', '') == 'file':
                index = book.meta[id].get('index', '')
                indexfile = os.path.normpath(os.path.join(book.data_dir, index))
                targetfile = util.get_meta_refreshed_file(indexfile)
                if targetfile:
                    moz_icon_url = f'moz-icon://{quote(os.path.basename(targetfile))}?size=16'
                    yield Info('debug', f'Generated moz-icon URL for "{id}": "{moz_icon_url}"')
                    return moz_icon_url
            return ''

        file = book.get_icon_file(book.meta[id])

        if not file:
            return icon

        file_ci = os.path.normcase(file)

        # favicon cache should go to "icon" folder
        favicon_dir = os.path.join(book.tree_dir, 'favicon', '')
        if file_ci.startswith(os.path.normcase(favicon_dir)):
            subpath = os.path.relpath(file, favicon_dir)
            fdst = os.path.join(self.output, 'icon', subpath)
            self.icons_to_cache[file_ci] = fdst
            yield Info('debug', f'Created icon file mapping (cache): "{file}" => "{fdst}"')
            return f'resource://scrapbook/icon/{pathname2url(subpath)}'

        # if inside data folder
        if file_ci.startswith(os.path.normcase(book.data_dir)):
            # if under an item folder, map to legacy item folder
            for path, oid in self.path_to_oid.items():
                if file_ci.startswith(path):
                    subpath = os.path.relpath(file, path)
                    fdst = os.path.join(self.output, 'data', oid, subpath)
                    self.icons_to_cache[file_ci] = fdst
                    yield Info('debug', f'Created icon file mapping (item): "{file}" => "{fdst}"')
                    return f'resource://scrapbook/data/{oid}/{pathname2url(subpath)}'

            # otherwise, map to sub-data-directory path
            subpath = os.path.relpath(file, book.data_dir)
            fdst = os.path.join(self.output, 'data', subpath)
            self.icons_to_cache[file_ci] = fdst
            yield Info('debug', f'Created icon file mapping (data): "{file}" => "{fdst}"')
            return f'resource://scrapbook/data/{pathname2url(subpath)}'

        # record icons outside of "data" folder to copy later
        subpath = os.path.relpath(file, book.top_dir)
        fdst = os.path.join(self.output, subpath)
        self.icons_to_cache[file_ci] = fdst
        yield Info('debug', f'Created icon file mapping: "{file}" => "{fdst}"')
        return f'resource://scrapbook/{pathname2url(subpath)}'

    def _copy_icon_files(self):
        for fsrc, fdst in self.icons_to_cache.items():
            yield Info('debug', f'Copying icon "{fsrc}" => "{fdst}"')
            if not os.path.exists(fdst):
                os.makedirs(os.path.dirname(fdst), exist_ok=True)
                shutil.copy2(fsrc, fdst)


class ConvertHtmlFile(HtmlRewriter):
    def rewrite(self, markups):
        markups, _ = self.convert(markups)
        return markups

    def convert(self, markups, start=0, endtag=None):
        rv = []
        i = start
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if markup.type == 'starttag':
                # pass-through special context tags
                if markup.tag in ('template', 'svg', 'math'):
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023
                    for j in range(i, iend + 1):
                        rv.append(markups[j])
                    i = iend + 1
                    continue

                # handle annotations
                type = markup.getattr('data-scrapbook-elem')

                # linemarker / inline
                if type == 'linemarker':
                    id = markup.getattr('data-scrapbook-id')
                    title = markup.getattr('title')
                    style = markup.getattr('style')

                    tag = 'span'
                    attrs = {
                        'data-sb-id': id,
                        'data-sb-obj': 'inline' if title else 'linemarker',
                        'class': ['scrapbook-inline' if title else 'linemarker-marked-line'],
                    }

                    if style:
                        attrs['style'] = style

                    if title:
                        attrs['title'] = title

                    attrs['class'] = ' '.join(attrs['class'])
                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='starttag',
                        tag=tag,
                        attrs=attrs,
                    ))

                    _rv, _i = self.convert(markups, i + 1, markup.endtag)
                    rv.extend(_rv)
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='endtag',
                        tag=tag,
                    ))

                    i = _i + 1
                    self.changed = True
                    continue

                # freenote / sticky
                elif type == 'sticky' and 'styled' in markup.classes:
                    is_relative = 'relative' in markup.classes
                    is_plaintext = 'plaintext' in markup.classes

                    tag = 'div'
                    attrs = {
                        'data-sb-obj': 'freenote',
                        'style': None,
                    }

                    # style
                    attrs['style'] = ' '.join([
                        'cursor: help;',
                        'overflow: visible;',
                        ('margin: 16px auto; ' if is_relative else '') + 'border: 1px solid #CCCCCC;',
                        'border-top-width: 12px;',
                        'background: #FAFFFA;',
                        'opacity: 0.95;',
                        'padding: 0px;',
                        'z-index: 500000;',
                        'text-align: start;',
                        'font-size: small;',
                        'line-height: 1.2em;',
                        'word-wrap: break-word;',
                        f'position: {"static" if is_relative else "absolute"};',
                    ])

                    style = markup.getattr('style')
                    if style:
                        attrs['style'] += ' ' + style

                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='starttag',
                        tag=tag,
                        attrs=attrs,
                    ))

                    if is_plaintext:
                        iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023

                        last_child_i = i
                        for j in self.iterfind(markups, lambda x: x.type == 'endtag' and x != markup.endtag, i + 1, markup.endtag):  # noqa: B023
                            last_child_i = j
                        text = ''.join(str(d) for d in markups[last_child_i + 1:iend] if d.type == 'data')

                        for line in REGEX_LINEFEED.split(text):
                            rv.append(Markup(
                                is_xhtml=self.is_xhtml,
                                type='data',
                                data=line,
                                convert_charrefs=False,
                            ))
                            rv.append(MarkupTag(
                                is_xhtml=self.is_xhtml,
                                type='starttag',
                                tag='br',
                                attrs=[],
                                is_self_end=self.is_xhtml,
                            ))
                        rv.pop()  # pop an extra <br>

                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='endtag',
                            tag=tag,
                        ))
                        i = iend + 1

                    else:
                        _rv, _i = self.convert(markups, i + 1, markup.endtag)
                        rv.extend(_rv)
                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='endtag',
                            tag=tag,
                        ))

                        i = _i + 1

                    self.changed = True
                    continue

                # block-comment
                elif (type == 'sticky'
                      and 'styled' not in markup.classes
                      and 'relative' in markup.classes
                      and 'plaintext' in markup.classes
                      ):
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023

                    tag = 'div'
                    attrs = {
                        'class': ['scrapbook-block-comment'],
                        'style': markup.getattr('style'),
                    }

                    last_child_i = i
                    for j in self.iterfind(markups, lambda x: x.type == 'endtag' and x != markup.endtag, i + 1, markup.endtag):  # noqa: B023
                        last_child_i = j
                    text = ''.join(str(d) for d in markups[last_child_i + 1:iend] if d.type == 'data')

                    attrs['class'] = ' '.join(attrs['class'])
                    attrs = [(a, v) for a, v in attrs.items() if v]
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='starttag',
                        tag=tag,
                        attrs=attrs,
                    ))
                    rv.append(Markup(
                        is_xhtml=self.is_xhtml,
                        type='data',
                        data=text,
                        convert_charrefs=False,
                    ))
                    rv.append(MarkupTag(
                        is_xhtml=self.is_xhtml,
                        type='endtag',
                        tag=tag,
                    ))

                    i = iend + 1
                    self.changed = True
                    continue

                # remove WebScrapBook-specific elements
                elif type in ('annotation-css', 'annotation-loader'):
                    iend = self.find(markups, lambda x: x == markup.endtag, i + 1, markup.endtag)  # noqa: B023
                    i = iend + 1
                    self.changed = True
                    continue

                else:
                    markup_changed = False
                    for j, attr_value in enumerate(markup.attrs):
                        attr, value = attr_value

                        # convert data-scrapbook-elem
                        if attr == 'data-scrapbook-elem':
                            markup.attrs[j] = ('data-sb-obj', type)
                            markup_changed = True

                        # convert id attribute
                        elif attr == 'data-scrapbook-id':
                            markup.attrs[j] = ('data-sb-id', value)
                            markup_changed = True

                    if markup_changed:
                        rv.append(MarkupTag(
                            is_xhtml=self.is_xhtml,
                            type='starttag',
                            tag=markup.tag,
                            attrs=markup.attrs,
                            is_self_end=markup.is_self_end,
                        ))

                        i += 1
                        self.changed = True
                        continue

            elif markup.type == 'endtag':
                if endtag is not None:
                    if markup == endtag:
                        break

            rv.append(markup)
            i += 1

        return rv, i


def run(input, output, book_id='', no_data_files=False):
    start = time.time()
    yield Info('info', 'conversion mode: WebScrapBook --> ScrapBook')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', f'book: "{book_id}"')
    yield Info('info', f'no-data-files: {no_data_files}')
    yield Info('info', '')

    try:
        conv = Converter(input, output, book_id=book_id, no_data_files=no_data_files)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import os
import shutil
import time
import traceback
from datetime import datetime, timedelta, timezone
from urllib.parse import quote

from ... import WSB_DIR, util
from ...util import Info
from ..host import Host
from ..indexer import Indexer

DEFAULT_DATA_FOLDER_SUFFIXES = ['.files', '_files']


class Converter:
    def __init__(self, input, output, *,
                 data_folder_suffixes=None,
                 preserve_filename=True,
                 handle_ie_meta=True,
                 handle_singlefile_meta=True,
                 handle_savepagewe_meta=True,
                 handle_maoxian_meta=True,
                 ):
        self.input = os.path.abspath(input)
        self.output = os.path.abspath(output)
        self.data_folder_suffixes = (
            [k for k in dict.fromkeys(os.path.normcase(s.strip()) for s in data_folder_suffixes) if k]
            if data_folder_suffixes is not None else DEFAULT_DATA_FOLDER_SUFFIXES)
        self.preserve_filename = preserve_filename
        self.handle_ie_meta = handle_ie_meta
        self.handle_singlefile_meta = handle_singlefile_meta
        self.handle_savepagewe_meta = handle_savepagewe_meta
        self.handle_maoxian_meta = handle_maoxian_meta
        self.wsb_dir = os.path.join(self.input, WSB_DIR)
        self.host = None
        self.book = None

    def run(self):
        self.host = Host(self.output)
        self.book = self.host.books['']

        self.book.load_meta_files()
        self.book.load_toc_files()

        yield Info('info', 'Inspecting data files...')
        paths = []
        yield from self._inspect_data_dir(self.input, paths)

        yield Info('info', 'Saving tree files...')
        self.book.save_meta_files()
        self.book.save_toc_files()

    def _inspect_data_dir(self, data_dir, paths):
        yield Info('debug', f'Inspecting directory "{data_dir}"')

        if not os.path.samefile(data_dir, self.input):
            # add data_dir as an item if index.html exists
            index = os.path.join(data_dir, 'index.html')
            if os.path.isfile(index):
                yield from self._index_entry(data_dir, paths)
                return

            # create folder item
            basename = os.path.basename(data_dir)
            basename, ext = os.path.splitext(basename)
            if ext.lower() == '.htd':
                ext = ''

            id = self._generate_unique_id()
            meta = self.book.DEFAULT_META.copy()
            meta.update({
                'type': 'folder',
                'title': basename + ext,
                'create': id,
                'modify': id,
            })
            for key in list(meta):
                if meta[key] is None:
                    del meta[key]
            self.book.meta[id] = meta

            parent_id = paths[-1]
            self.book.toc.setdefault(parent_id, []).append(id)
            yield Info('info', f'Generated folder item "{id}" under "{parent_id}"')
        else:
            id = 'root'

        try:
            entries = os.scandir(data_dir)
        except FileNotFoundError:
            return
        except OSError as exc:
            yield Info('error', f'Failed to scan folder "{self.book.get_subpath(exc.filename)}": {exc.strerror}', exc=exc)
            return

        entries_to_handle = set()
        entries_to_exclude = set()
        with entries as entries:
            for entry in entries:
                if os.path.normcase(entry.path) == os.path.normcase(self.wsb_dir):
                    yield Info('debug', f'Skipped special "{self.book.get_subpath(entry)}"')
                    continue

                if entry.is_dir():
                    entries_to_handle.add(entry)

                elif entry.is_file():
                    if self._get_index_path_key(entry) not in entries_to_exclude:
                        entries_to_handle.add(entry)

                        if util.is_html(entry.path):
                            basename, _ = os.path.splitext(entry.name)
                            for suffix in self.data_folder_suffixes:
                                p = self._get_index_path_key(os.path.join(data_dir, f'{basename}{suffix}'))
                                yield Info('debug', f'Excluding "{p}" from index finding')
                                entries_to_exclude.add(p)

        paths.append(id)

        for entry in sorted(entries_to_handle, key=lambda x: x.path):
            if self._get_index_path_key(entry) in entries_to_exclude:
                continue

            if entry.is_dir():
                yield from self._inspect_data_dir(entry.path, paths)
            elif entry.is_file():
                yield from self._index_entry(entry.path, paths)

        paths.pop()

    def _index_entry(self, entry, paths):
        yield Info('debug', f'Generating item for "{entry}"...')

        basename = os.path.basename(entry)
        _, ext = os.path.splitext(entry)
        ext = ext.lower()
        if ext == '.htd':
            ext = ''

        # generate a unique ID
        id = self._generate_unique_id(ext)

        # copy data files
        supporting_folder = self._get_supporting_folder(entry)
        if (supporting_folder or (
            self.preserve_filename and os.path.isfile(entry) and not util.is_archive(entry)
        )):
            dst_dir = os.path.join(self.book.data_dir, id)
            os.makedirs(dst_dir, exist_ok=True)

            src = entry
            dst = os.path.join(dst_dir, basename)
            yield Info('info', f'Copying data file: "{src}" => "{dst}"')
            try:
                shutil.copy2(src, dst)
            except OSError as exc:
                yield Info('error', f'Failed to copy data file "{entry}": {exc.strerror}', exc=exc)

            index_file = os.path.join(dst_dir, 'index.html')
            if ext in self.book.ITEM_INDEX_ALLOWED_EXT:
                # copy entry to index.html for the indexer to retrieve original metadata
                try:
                    shutil.copy2(src, index_file)
                except OSError as exc:
                    yield Info('error', f'Failed to copy data file "{entry}": {exc.strerror}', exc=exc)

                if supporting_folder:
                    src = supporting_folder
                    dst = os.path.join(dst_dir, os.path.basename(supporting_folder))
                    yield Info('info', f'Copying data folder: "{src}" => "{dst}"')
                    try:
                        shutil.copytree(src, dst)
                    except OSError as exc:
                        yield Info('error', f'Failed to copy data folder "{entry}": {exc.strerror}', exc=exc)

                # generate meta
                indexer = Indexer(
                    self.book,
                    handle_ie_meta=self.handle_ie_meta,
                    handle_singlefile_meta=self.handle_singlefile_meta,
                    handle_savepagewe_meta=self.handle_savepagewe_meta,
                    handle_maoxian_meta=self.handle_maoxian_meta,
                )
                indexed = yield from indexer.run([index_file])

                if os.path.normcase(basename) != os.path.normcase('index.html'):
                    with open(index_file, 'w', encoding='UTF-8') as fh:
                        fh.write(f'<!DOCTYPE html><meta charset="UTF-8"><meta http-equiv="refresh" content="0; url={quote(basename)}">')
            else:
                # generate new index.html (with same file time) for the indexer
                with open(index_file, 'w', encoding='UTF-8') as fh:
                    fh.write(f'<!DOCTYPE html><meta charset="UTF-8"><meta http-equiv="refresh" content="0; url={quote(basename)}">')
                st = os.stat(entry)
                os.utime(index_file, (st.st_atime, st.st_mtime))

                # generate meta
                indexer = Indexer(
                    self.book,
                    handle_ie_meta=self.handle_ie_meta,
                    handle_singlefile_meta=self.handle_singlefile_meta,
                    handle_savepagewe_meta=self.handle_savepagewe_meta,
                    handle_maoxian_meta=self.handle_maoxian_meta,
                )
                indexed = yield from indexer.run([index_file])

        else:
            src = entry
            dst = os.path.join(self.book.data_dir, id + ext)
            yield Info('info', f'Copying data files: "{src}" => "{dst}"')
            try:
                try:
                    shutil.copytree(src, dst)
                except NotADirectoryError:
                    shutil.copy2(src, dst)
            except OSError as exc:
                yield Info('error', f'Failed to copy data files for "{entry}": {exc.strerror}', exc=exc)

            index_file = os.path.join(dst, 'index.html') if os.path.isdir(entry) else dst

            # generate meta
            indexer = Indexer(
                self.book,
                handle_ie_meta=self.handle_ie_meta,
                handle_singlefile_meta=self.handle_singlefile_meta,
                handle_savepagewe_meta=self.handle_savepagewe_meta,
                handle_maoxian_meta=self.handle_maoxian_meta,
            )
            indexed = yield from indexer.run([index_file])

        for id in indexed:
            # special handle of metadata
            meta = self.book.meta[id]
            if not meta.get('title'):
                meta['title'] = basename

            if os.path.isfile(entry) and ext not in self.book.ITEM_INDEX_ALLOWED_EXT:
                meta['type'] = 'file'

            # add to parent
            parent_id = paths[-1]
            self.book.toc.setdefault(parent_id, []).append(id)
            yield Info('info', f'Appended item "{id}" under "{parent_id}"')

    def _get_index_path_key(self, path):
        return self.book.get_subpath(os.path.normcase(path))

    def _get_supporting_folder(self, file):
        if os.path.isfile(file) and util.is_html(file):
            base = os.path.splitext(file)[0]

            for suffix in self.data_folder_suffixes:
                supporting_folder = f'{base}{suffix}'
                if os.path.isdir(supporting_folder):
                    return supporting_folder

        return None

    def _generate_unique_id(self, ext=''):
        dt = datetime.now(timezone.utc)
        id = util.datetime_to_id(dt)
        while id in self.book.meta or os.path.lexists(os.path.join(self.book.data_dir, id + ext)):
            dt += timedelta(milliseconds=1)
            id = util.datetime_to_id(dt)
        return id


def run(input, output, *,
        data_folder_suffixes=None,
        no_preserve_filename=False,
        ignore_ie_meta=False,
        ignore_singlefile_meta=False,
        ignore_savepagewe_meta=False,
        ignore_maoxian_meta=False,
        ):
    start = time.time()
    yield Info('info', 'conversion mode: hierarchical files --> WebScrapBook')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', f'data_folder_suffixes: {DEFAULT_DATA_FOLDER_SUFFIXES if data_folder_suffixes is None else data_folder_suffixes}')
    yield Info('info', f'no preserve filename: {no_preserve_filename}')
    yield Info('info', f'ignore IE meta: {ignore_ie_meta}')
    yield Info('info', f'ignore SingleFile meta: {ignore_singlefile_meta}')
    yield Info('info', f'ignore Save Page WE meta: {ignore_savepagewe_meta}')
    yield Info('info', f'ignore MaoXian web clipper meta: {ignore_maoxian_meta}')
    yield Info('info', '')

    try:
        conv = Converter(
            input, output,
            data_folder_suffixes=data_folder_suffixes,
            preserve_filename=not no_preserve_filename,
            handle_ie_meta=not ignore_ie_meta,
            handle_singlefile_meta=not ignore_singlefile_meta,
            handle_savepagewe_meta=not ignore_savepagewe_meta,
            handle_maoxian_meta=not ignore_maoxian_meta,
        )
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import json
import os
import time
import traceback
import zipfile
from contextlib import nullcontext
from datetime import datetime, timedelta, timezone

from .. import util
from ..util import Info
from .host import Host


class Exporter():
    """Main class for generating exports.
    """
    def __init__(self, output, book, *, singleton=False):
        self.output = output
        self.book = book
        self.singleton = singleton

        self.recycle = None
        self.used_ts = None
        self.map_id_to_eid = None

    def run(self, item_ids=None, recursive=False):
        book = self.book
        self.book.load_meta_files()
        self.book.load_toc_files()

        os.makedirs(self.output, exist_ok=True)

        self.recycle = set(book.toc.get('recycle', []))
        self.used_ts = set()
        self.map_id_to_eid = {}

        if item_ids:
            id_pool = {id for id in item_ids if id in book.meta and id not in self.recycle}

            # add descendant id if recursive mode
            if recursive:
                for id in list(id_pool):
                    for desc_id in self._iter_child_items(id, [id]):
                        id_pool.add(desc_id)
                        yield Info('debug', f'Included descendant item of "{id}": "{desc_id}"')
        else:
            id_pool = {id for id in book.meta if id not in self.recycle}

        id_chain = ['root']
        for id in self._iter_child_items('root', id_chain):
            if id in id_pool:
                yield from self._export_item(id, id_chain)

        id_chain = ['hidden']
        for id in self._iter_child_items('hidden', id_chain):
            if id in id_pool:
                yield from self._export_item(id, id_chain)

    def _iter_child_items(self, id, id_chain):
        """Generate descendant items for the item of id.
        """
        for ref_id in self.book.toc.get(id, []):
            yield ref_id

            # do not export children of a circular item
            if ref_id not in id_chain:
                id_chain.append(ref_id)
                yield from self._iter_child_items(ref_id, id_chain)
                id_chain.pop()

    def _export_item(self, id, id_chain):
        if id in self.map_id_to_eid:
            if self.singleton:
                yield Info('debug', f'Skipped exporting item "{id}" (singleton mode)')
                return

        yield Info('debug', f'Exporting item "{id}"')
        try:
            yield from self._export_item_internal(id, id_chain)
        except Exception as exc:
            # unexpected error
            traceback.print_exc()
            yield Info('error', f'Failed to export "{id}": {exc}', exc=exc)

    def _export_item_internal(self, id, id_chain):
        meta = self.book.meta[id]
        index = meta.get('index', '')

        # generate a unique timestamp as prefix
        ts = datetime.now(timezone.utc)
        ets = util.datetime_to_id(ts)
        while ets in self.used_ts:
            ts += timedelta(milliseconds=1)
            ets = util.datetime_to_id(ts)
        self.used_ts.add(ets)

        # setup an export id (eid), which is unique and is same among all
        # occurrences of the same id, to the ets of the first occurrence
        eid = self.map_id_to_eid.setdefault(id, ets)

        # generate a unique timestamp prefix
        basename = ets + '-' + meta.get('title', meta.get('id', ''))
        basename = util.crop(util.validate_filename(basename), 128)

        # generate a unique filename
        i = 0
        dst = os.path.join(self.output, f'{basename}.wsba')
        while os.path.lexists(dst):
            i += 1
            dst = os.path.join(self.output, f'{basename}-{i}.wsba')

        yield Info('info', f'Exporting "{id}" to "{os.path.basename(dst)}"')
        parents = [{'id': id, 'title': self.book.meta.get(id, {}).get('title', '')} for id in id_chain]
        meta_data = {'id': id, **meta}
        export_data = {
            'version': 1,
            'id': eid,
            'timestamp': ets,
            'timezone': datetime.now().astimezone().utcoffset().total_seconds(),
            'path': parents,
        }
        with zipfile.ZipFile(dst, 'w') as zh:
            zh.writestr('meta.json', json.dumps(meta_data, ensure_ascii=False, indent=2))
            zh.writestr('export.json', json.dumps(export_data, ensure_ascii=False, indent=2))

            # include data file(s)
            if index:
                zh.writestr('data/', '')
                src = os.path.join(self.book.data_dir, os.path.dirname(index) if index.endswith('/index.html') else index)
                yield Info('debug', f'Saving data files for "{id}": "{self.book.get_subpath(src)}"')
                util.zip_compress(zh, src, f'data/{os.path.basename(src)}')

            # include favicon cache
            iconfile = self.book.get_icon_file(meta)
            if not iconfile:
                return

            favicon_dir = os.path.join(self.book.tree_dir, 'favicon', '')
            if not os.path.normcase(iconfile).startswith(os.path.normcase(favicon_dir)):
                return

            zh.writestr('favicon/', '')
            util.zip_compress(zh, iconfile, f'favicon/{os.path.basename(iconfile)}')


def run(root, output, book_id='', item_ids=None, *, recursive=False, singleton=False,
        config=None, no_lock=False):
    start = time.time()

    host = Host(root, config)

    # Fail for invalid book ID
    if book_id not in host.books:
        yield Info('error', f'Invalid book "{book_id}".')
        return

    yield Info('debug', f'Loading book "{book_id}".')

    try:
        book = host.books[book_id]

        if book.no_tree:
            yield Info('error', f'Unable to export book "{book_id}" ("{book.name}") (no_tree).')
            return

        yield Info('info', f'Exporting from book "{book_id}" ({book.name}).')
        lh = nullcontext() if no_lock else book.get_tree_lock().acquire()
        with lh:
            generator = Exporter(output, book, singleton=singleton)
            yield from generator.run(item_ids, recursive)

    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import copy
import html
import os
import shutil
import time
import traceback
import zipfile
from datetime import datetime, timezone
from email.utils import format_datetime
from urllib.parse import quote

from ... import util
from ...util import Info
from ..host import Host
from ..indexer import FavIconCacher, SingleHtmlConverter, UnSingleHtmlConverter


class Converter:
    def __init__(self, input, output, book_ids=None, item_ids=None, types=None, format=None):
        self.input = input
        self.output = output
        self.book_ids = book_ids
        self.item_ids = item_ids
        self.types = set(types) if types else {}
        self.format = format

    def run(self):
        if self.input != self.output:
            yield Info('info', 'Copying files...')
            self._copy_files()

        yield Info('info', 'Applying conversion...')
        host = Host(self.output)

        # handle all books if none specified
        for book_id in self.book_ids or host.books:
            try:
                book = host.books[book_id]
            except KeyError:
                # skip invalid book ID
                yield Info('warn', f'Skipped invalid book "{book_id}".')
                continue

            yield Info('info', f'Handling book "{book_id}"...')
            book.load_meta_files()

            book_meta_orig = copy.deepcopy(book.meta)

            # handle all items if none specified
            for id in self.item_ids or book.meta:
                if id not in book.meta:
                    # skip invalid item ID
                    yield Info('debug', f'Skipped invalid item "{id}".')
                    continue

                type = book.meta[id].get('type', '')
                if type not in self.types:
                    yield Info('debug', f'Skipped item "{id}": type="{type}"')
                    continue

                yield Info('debug', f'Checking "{id}"...')

                if self.format:
                    try:
                        try:
                            yield from self._convert_item_format(book, id)
                        except OSError as exc:
                            raise RuntimeError(exc.strerror) from exc
                    except Exception as exc:
                        traceback.print_exc()
                        yield Info('error', f'Failed to convert "{id}": {exc}', exc=exc)

            # update files
            if book.meta != book_meta_orig:
                yield Info('info', 'Saving changed meta files...')
                book.save_meta_files()

    def _copy_files(self):
        with os.scandir(self.input) as dirs:
            for src in dirs:
                dst = os.path.join(self.output, src.name)
                try:
                    shutil.copytree(src, dst)
                except NotADirectoryError:
                    shutil.copy2(src, dst)

    def _convert_item_format(self, book, id):
        meta = book.meta[id]
        index = meta.get('index')

        if not index:
            yield Info('debug', f'Skipped "{id}": no index')
            return

        if index.endswith('/index.html'):
            format = 'folder'
        elif util.is_htz(index):
            format = 'htz'
        elif util.is_maff(index):
            format = 'maff'
        else:
            format = 'single_file'

        if format == self.format:
            yield Info('debug', f'Skipped "{id}": same format')
            return

        if format == 'folder':
            indexbase = index[:-11]
            fsrc = os.path.normpath(os.path.join(book.data_dir, indexbase))
            indexdir = os.path.normpath(os.path.join(book.data_dir, indexbase + '.' + util.datetime_to_id()))
            shutil.copytree(fsrc, indexdir)
            yield from self._cache_favicon(book, id)
        elif format == 'htz':
            fsrc = os.path.normpath(os.path.join(book.data_dir, index))
            indexbase = index[:-4]
            indexdir = os.path.normpath(os.path.join(book.data_dir, indexbase + '.' + util.datetime_to_id()))
            util.zip_extract(fsrc, indexdir)
        elif format == 'maff':
            fsrc = os.path.normpath(os.path.join(book.data_dir, index))
            indexbase = index[:-5]
            indexdir = os.path.normpath(os.path.join(book.data_dir, indexbase + '.' + util.datetime_to_id()))

            maff_info = next(iter(util.get_maff_pages(fsrc)), None)
            if not maff_info:
                yield Info('debug', f'Skipping "{id}": no valid index page in MAFF')
            subpath, _, _ = maff_info.indexfilename.partition('/')

            util.zip_extract(fsrc, indexdir, subpath)

            rdf_file = os.path.join(indexdir, 'index.rdf')
            try:
                os.remove(rdf_file)
            except FileNotFoundError:
                pass
        else:
            fsrc = os.path.normpath(os.path.join(book.data_dir, index))
            indexbase, ext = os.path.splitext(index)
            indexdir = os.path.normpath(os.path.join(book.data_dir, indexbase + '.' + util.datetime_to_id()))

            os.makedirs(indexdir)
            indexfile = os.path.join(indexdir, 'index.html')
            if util.is_html(fsrc) and not util.is_xhtml(fsrc):
                mainfile = indexfile
                shutil.copy2(fsrc, mainfile)
            else:
                basename = os.path.basename(index)
                mainfile = os.path.join(indexdir, basename)
                shutil.copy2(fsrc, mainfile)
                with open(indexfile, 'w', encoding='UTF-8', newline='\n') as fh:
                    fh.write(f'<!DOCTYPE html><meta charset="UTF-8"><meta http-equiv="refresh" content="0; url={quote(basename)}">')

            if util.is_html(mainfile) or util.is_svg(mainfile):
                conv = UnSingleHtmlConverter(mainfile)
                content = conv.run()
                with open(mainfile, 'w', encoding=conv.encoding, newline='') as fh:
                    fh.write(content)

            shutil.copystat(fsrc, indexfile)

        try:
            if self.format == 'folder':
                fdst = os.path.normpath(os.path.join(book.data_dir, indexbase))
                yield Info('info', f'Converting "{id}": "{book.get_subpath(fsrc)}" => "{book.get_subpath(fdst)}" ...')

                if os.path.lexists(fdst):
                    yield Info('error', f'Failed to convert "{id}": target "{book.get_subpath(fdst)}" already exists.')
                    return

                shutil.move(indexdir, fdst)

                # adjust icon path to fit the new index file
                iconfile = book.get_icon_file(meta)
                if iconfile:
                    meta['icon'] = util.get_relative_url(iconfile, fdst, path_is_dir=False, start_is_dir=True)

                meta['index'] = indexbase + '/index.html'

            elif self.format == 'htz':
                fdst = os.path.normpath(os.path.join(book.data_dir, indexbase + '.htz'))
                yield Info('info', f'Converting "{id}": "{book.get_subpath(fsrc)}" => "{book.get_subpath(fdst)}" ...')

                if os.path.lexists(fdst):
                    yield Info('error', f'Failed to convert "{id}": target "{book.get_subpath(fdst)}" already exists.')
                    return

                util.zip_compress(fdst, indexdir, '')
                shutil.copystat(os.path.join(indexdir, 'index.html'), fdst)

                # adjust icon path to fit the new index file
                iconfile = book.get_icon_file(meta)
                if iconfile:
                    meta['icon'] = util.get_relative_url(iconfile, fdst, path_is_dir=False, start_is_dir=False)

                meta['index'] = indexbase + '.htz'

            elif self.format == 'maff':
                fdst = os.path.normpath(os.path.join(book.data_dir, indexbase + '.maff'))
                yield Info('info', f'Converting "{id}": "{book.get_subpath(fsrc)}" => "{book.get_subpath(fdst)}" ...')

                rdf_file = os.path.join(indexdir, 'index.rdf')
                if os.path.lexists(rdf_file):
                    yield Info('error', f'Failed to convert "{id}": index.rdf file already exists.')
                    return

                if os.path.lexists(fdst):
                    yield Info('error', f'Failed to convert "{id}": target "{book.get_subpath(fdst)}" already exists.')
                    return

                subpath = id if util.id_to_datetime(id) else util.datetime_to_id()
                util.zip_compress(fdst, indexdir, subpath)

                rdf_content = self._generate_index_rdf(book, id)
                with zipfile.ZipFile(fdst, 'a') as zh:
                    zh.writestr(
                        f'{subpath}/index.rdf', rdf_content,
                        **util.zip_compression_params(mimetype='application/rdf+xml')
                    )

                shutil.copystat(os.path.join(indexdir, 'index.html'), fdst)

                # adjust icon path to fit the new index file
                iconfile = book.get_icon_file(meta)
                if iconfile:
                    meta['icon'] = util.get_relative_url(iconfile, fdst, path_is_dir=False, start_is_dir=False)

                meta['index'] = indexbase + '.maff'

            elif self.format == 'single_file':
                file = os.path.join(indexdir, 'index.html')
                file = util.get_meta_refreshed_file(file) or file

                if util.is_xhtml(file):
                    ext = '.xhtml'
                elif util.is_html(file):
                    ext = '.html'
                elif util.is_svg(file):
                    ext = '.svg'
                else:
                    _, ext = os.path.splitext(file)

                # special handling to prevent named "index.html"
                if indexbase == 'index' and ext == '.html':
                    indexbase = 'index_'

                fdst = os.path.normpath(os.path.join(book.data_dir, indexbase + ext))
                yield Info('info', f'Converting "{id}": "{book.get_subpath(fsrc)}" => "{book.get_subpath(fdst)}" ...')

                if os.path.lexists(fdst):
                    yield Info('error', f'Failed to convert "{id}": target "{book.get_subpath(fdst)}" already exists.')
                    return

                if util.is_html(file) or util.is_svg(file):
                    conv = SingleHtmlConverter(file)
                    content = conv.run()
                    with open(fdst, 'w', encoding=conv.encoding, newline='') as fh:
                        fh.write(content)
                    shutil.copystat(file, fdst)
                else:
                    shutil.copy2(file, fdst)

                if meta.get('icon'):
                    iconfile = book.get_icon_file(meta)
                    meta['icon'] = util.get_relative_url(iconfile, fdst, path_is_dir=False, start_is_dir=False)
                meta['index'] = indexbase + ext

            try:
                shutil.rmtree(fsrc)
            except NotADirectoryError:
                os.remove(fsrc)
        finally:
            try:
                shutil.rmtree(indexdir)
            except FileNotFoundError:
                pass

    def _cache_favicon(self, book, id):
        generator = FavIconCacher(book, cache_archive=True, cache_file=True)
        yield from generator.run([id])

    def _generate_index_rdf(self, book, id):
        meta = book.meta[id]
        dt = util.id_to_datetime(meta.get('create', ''))
        dt = dt.astimezone() if dt else datetime.now(timezone.utc)
        return f"""\
<?xml version="1.0"?>
<RDF:RDF xmlns:MAF="http://maf.mozdev.org/metadata/rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
<RDF:Description RDF:about="urn:root">
  <MAF:originalurl RDF:resource="{html.escape(meta.get('source', ''))}"/>
  <MAF:title RDF:resource="{html.escape(meta.get('title', ''))}"/>
  <MAF:archivetime RDF:resource="{format_datetime(dt)}"/>
  <MAF:indexfilename RDF:resource="index.html"/>
  <MAF:charset RDF:resource="{html.escape(meta.get('charset') or 'UTF-8')}"/>
</RDF:Description>
</RDF:RDF>
"""


def run(input, output, book_ids=None, item_ids=None, types=None, format=None):
    start = time.time()
    book_ids_text = ', '.join(f'"{id}"' for id in book_ids) if book_ids else 'all'
    item_ids_text = ', '.join(f'"{id}"' for id in item_ids) if item_ids else 'all'
    yield Info('info', 'converting items:')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output) if output is not None else "(in-place)"}')
    yield Info('info', f'book(s): {book_ids_text}')
    yield Info('info', f'item(s): {item_ids_text}')
    yield Info('info', f'types: {types}')
    yield Info('info', f'format: {format}')
    yield Info('info', '')

    if output is None:
        output = input

    try:
        conv = Converter(input, output, book_ids=book_ids, item_ids=item_ids, types=types, format=format)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



#!/usr/bin/env python3
from inspect import cleandoc

from setuptools import find_packages, setup

import webscrapbook

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='webscrapbook',
    version=webscrapbook.__version__,
    description=cleandoc(webscrapbook.__doc__),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=webscrapbook.__author__,
    author_email=webscrapbook.__author_email__,
    url=webscrapbook.__homepage__,
    license=webscrapbook.__license__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
    ],
    python_requires='~=3.7',
    install_requires=[
        'flask >= 1.1',
        'werkzeug >= 1.0.0',
        'jinja2 >= 2.10.1',
        'lxml >= 4.0',
        'commonmark >= 0.8',
    ],
    extras_require={
        'adhoc_ssl': ['cryptography'],
        'dev': [
            'flake8 >= 4.0',
            'pep8-naming >= 0.13.2',
            'flake8-comprehensions >= 3.7',
            'flake8-string-format >= 0.3',
            'flake8-quotes >= 3.0',
            'flake8-bugbear >= 22.0',
            'flake8-isort >= 4.2',
            'isort >= 5.5',
        ],
    },
    packages=find_packages(exclude=['tests']),
    package_data={
        'webscrapbook': [
            'resources/*.*',
            'themes/default/static/*.*',
            'themes/default/templates/*.*',
            'themes/default/locales/*/*.py',
        ],
    },
    entry_points={
        'console_scripts': [
            'webscrapbook = webscrapbook.cli:main',
            'wsb = webscrapbook.cli:main',
            'wsbview = webscrapbook.cli:view',
        ],
    },
)



import copy
import json
import os
import re
import shutil
import time
import traceback
import uuid
import zipfile
from contextlib import nullcontext
from datetime import datetime, timedelta, timezone

from .. import util
from ..util import Info
from .book import Book
from .host import Host

REGEX_TARGET_FILENAME_FORMATTER = re.compile(r'%([^%]*)%')


class Importer():
    """Main class for importing.
    """
    def __init__(self, book, *,
                 target_id=None,
                 target_index=None,
                 target_filename=None,
                 rebuild_folders=False,
                 prune=False,
                 resolve_id_used='skip',  # skip, replace, new
                 ):
        self.book = book
        self.target_id = target_id or 'root'
        self.target_index = target_index
        self.target_filename = target_filename or '%ID%'
        self.rebuild_folders = rebuild_folders
        self.prune = prune
        self.resolve_id_used = resolve_id_used

        self.map_eid_to_id = None
        self.map_id_to_new_id = None

    def run(self, files=None):
        book = self.book
        self.book.load_meta_files()
        self.book.load_toc_files()

        self.map_eid_to_id = {}
        self.map_id_to_new_id = {}

        book_meta_orig = copy.deepcopy(book.meta)
        book_toc_orig = copy.deepcopy(book.toc)

        for file in files:
            if os.path.isdir(file):
                with os.scandir(file) as entries:
                    srcs = sorted(f.path for f in entries if util.is_wsba(f.path))
            elif os.path.isfile(file):
                if not util.is_wsba(file):
                    yield Info('warn', f'Skipped invalid file "{os.path.basename(file)}"')
                    continue
                srcs = [file]
            else:
                yield Info('error', f'Failed to import file "{os.path.basename(file)}": unable to access file')
                continue

            for src in srcs:
                try:
                    yield Info('debug', f'Importing file "{os.path.basename(src)}"')
                    id, eid, parent_id = yield from self._import_file(src)
                except RuntimeError as exc:
                    # intended raise to skip the import
                    yield Info('error', f'Failed to import file "{os.path.basename(src)}": {exc}', exc=exc)
                except Exception as exc:
                    # unexpected error
                    traceback.print_exc()
                    yield Info('error', f'Failed to import file "{os.path.basename(src)}": {exc}', exc=exc)
                else:
                    # finalize a successful import
                    yield Info('info', f'Imported "{id}" (under "{parent_id}")')
                    self.map_eid_to_id.setdefault(eid, id)
                    if self.prune:
                        yield Info('debug', f'Removing "{os.path.basename(src)}" (prune)')
                        os.remove(src)

        # update files
        if self.book.meta != book_meta_orig:
            yield Info('info', 'Saving changed meta files...')
            self.book.save_meta_files()

        if self.book.toc != book_toc_orig:
            yield Info('info', 'Saving changed TOC files...')
            self.book.save_toc_files()

    def generate_imported_filename(self, id, meta, export_info):
        """Generate an adequate filename (without file extension) for an
        importing item.
        """
        def date_formatter(id, pattern):
            if pattern == '':
                return id

            dt = util.id_to_datetime(id)

            if not dt:
                return ''

            if pattern == 'UTC_DATE':
                return f'{dt.year:04d}-{dt.month:02d}-{dt.day:02d}'

            if pattern == 'UTC_TIME':
                return f'{dt.hour:02d}-{dt.minute:02d}-{dt.second:02d}'

            if pattern == 'UTC_YEAR':
                return f'{dt.year:04d}'

            if pattern == 'UTC_MONTH':
                return f'{dt.month:02d}'

            if pattern == 'UTC_DAY':
                return f'{dt.day:02d}'

            if pattern == 'UTC_HOURS':
                return f'{dt.hour:02d}'

            if pattern == 'UTC_MINUTES':
                return f'{dt.minute:02d}'

            if pattern == 'UTC_SECONDS':
                return f'{dt.second:02d}'

            ldt = dt.astimezone()

            if pattern == 'DATE':
                return f'{ldt.year:04d}-{ldt.month:02d}-{ldt.day:02d}'

            if pattern == 'TIME':
                return f'{ldt.hour:04d}-{ldt.minute:02d}-{ldt.second:02d}'

            if pattern == 'YEAR':
                return f'{ldt.year:04d}'

            if pattern == 'MONTH':
                return f'{ldt.month:02d}'

            if pattern == 'DAY':
                return f'{ldt.day:02d}'

            if pattern == 'HOURS':
                return f'{ldt.hour:02d}'

            if pattern == 'MINUTES':
                return f'{ldt.minute:02d}'

            if pattern == 'SECONDS':
                return f'{ldt.second:02d}'

            return ''

        def formatter(m):
            key = m.group(1)

            if key == '':
                return '%'

            if key == 'ID':
                return id

            if key == 'EID':
                return export_info['id']

            if key == 'UUID':
                return str(uuid.uuid4())

            if key == 'TITLE':
                return meta.get('title', '')

            if key == 'SOURCE':
                return meta.get('source', '')

            key, _, pattern = key.partition(':')

            if key == 'CREATE':
                return str(date_formatter(meta.get('create', ''), pattern))

            if key == 'MODIFY':
                return str(date_formatter(meta.get('modify', ''), pattern))

            if key == 'EXPORT':
                return str(date_formatter(export_info['timestamp'], pattern))

            return ''

        filename = REGEX_TARGET_FILENAME_FORMATTER.sub(formatter, self.target_filename)
        filename = '/'.join(util.validate_filename(s) for s in filename.split('/'))
        return filename

    def _import_file(self, file):
        with zipfile.ZipFile(file) as zh:
            with zh.open('meta.json') as fh:
                meta = json.load(fh)

            with zh.open('export.json') as fh:
                export_info = json.load(fh)

            id = meta.pop('id')
            if id in Book.SPECIAL_ITEM_ID:
                raise RuntimeError(f'invalid ID "{id}"')

            # duplicated occurrence of a previously imported id
            # add to TOC only
            if self.map_eid_to_id.get(export_info['id']) == id:
                parent_id = yield from self._insert_to_toc(id, export_info)
                return (id, export_info['id'], parent_id)

            # do the import
            id = yield from self._import_meta_and_data(id, meta, zh, export_info)
            parent_id = yield from self._insert_to_toc(id, export_info)
            return (id, export_info['id'], parent_id)

    def _import_meta_and_data(self, id, meta, zh, export_info):
        """Import meta and data

        Returns:
            string: ID of the imported item
        """
        index = meta.get('index', '')

        if index:
            if index.endswith('/index.html'):
                src = f'data/{os.path.dirname(index)}'
            else:
                src = f'data/{index}'

            # determine normal copy dst
            _, ext = os.path.splitext(src)
            filename = self.generate_imported_filename(id, meta, export_info) + ext
            dst = os.path.normpath(os.path.join(self.book.data_dir, filename))
            meta['index'] = filename + ('/index.html' if index.endswith('/index.html') else '')

        # handle resolve cases if id exists
        # may overwrite id, dst, and meta['index']
        if id in self.book.meta:
            if self.resolve_id_used == 'skip':
                raise RuntimeError(f'ID "{id}" already exists')

            elif self.resolve_id_used == 'replace':
                index_old = self.book.meta[id].get('index', '')

                # replace only if index type matches
                if os.path.splitext(index)[1] != os.path.splitext(index_old)[1]:
                    raise RuntimeError('index type not match')

                if index_old.endswith('/index.html') != index.endswith('/index.html'):
                    raise RuntimeError('index type not match')

                yield Info('info', f'Force importing duplicated "{id}"...')

                if index:
                    # use original index
                    meta['index'] = index_old

                    # remove current index file or folder
                    if index.endswith('/index.html'):
                        dst = os.path.normpath(os.path.join(self.book.data_dir, os.path.dirname(index_old)))
                        try:
                            shutil.rmtree(dst)
                        except FileNotFoundError:
                            pass
                    else:
                        dst = os.path.normpath(os.path.join(self.book.data_dir, index_old))
                        try:
                            os.remove(dst)
                        except FileNotFoundError:
                            pass

            elif self.resolve_id_used == 'new':
                # generate a new unique id
                ts = datetime.now(timezone.utc)
                new_id = util.datetime_to_id(ts)
                while new_id in self.book.meta:
                    ts += timedelta(milliseconds=1)
                    new_id = util.datetime_to_id(ts)

                yield Info('info', f'Importing duplicated "{id}" as "{new_id}"...')
                self.map_id_to_new_id[id] = new_id
                id = new_id

                if index:
                    # overwrite dst and index
                    filename = self.generate_imported_filename(id, meta, export_info) + ext
                    dst = os.path.normpath(os.path.join(self.book.data_dir, filename))
                    meta['index'] = filename + ('/index.html' if index.endswith('/index.html') else '')

            else:
                raise RuntimeError(f'unknown resolve mode: "{self.resolve_id_used}"')

        # import data files
        if index:
            if os.path.lexists(dst):
                raise RuntimeError(f'file "{dst}" already exists')

            yield Info('debug', f'Extracting data files to "{self.book.get_subpath(dst)}"')
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            util.zip_extract(zh, dst, src, tzoffset=export_info['timezone'])

        # import favicon
        for f in zh.namelist():
            if f.startswith('favicon/') and not f.endswith('/'):
                try:
                    basename = os.path.basename(f)
                    dst = os.path.join(self.book.tree_dir, 'favicon', basename)
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                    util.zip_extract(zh, dst, f, tzoffset=export_info['timezone'])
                except FileExistsError:
                    yield Info('debug', f'Skipped existing favicon cache "{basename}"')
                else:
                    yield Info('info', f'Added favicon cache "{basename}"')

        self.book.meta[id] = meta
        return id

    def _insert_to_toc(self, id, export_info):
        """Insert the importing item to TOC

        Returns:
            string: ID of the parent the item is inserted under
        """
        if self.rebuild_folders:
            export_path = export_info['path']
            parent_id = export_path[-1]['id']
            parent_id = self.map_id_to_new_id.get(parent_id, parent_id)
        else:
            parent_id = self.target_id

        if parent_id in self.book.meta or parent_id in Book.SPECIAL_ITEM_ID:
            yield from self._insert_to_id(id, parent_id)
            return parent_id

        for i in reversed(range(len(export_path) - 1)):
            parent_id = export_path[i]['id']
            if parent_id in self.book.meta or parent_id in Book.SPECIAL_ITEM_ID:
                break
        else:
            # for a bad path data not starting from 'root'
            i = -1
            parent_id = 'root'

        for j in range(i + 1, len(export_path)):
            # generate a new unique id
            ts = datetime.now(timezone.utc)
            new_id = util.datetime_to_id(ts)
            while new_id in self.book.meta:
                ts += timedelta(milliseconds=1)
                new_id = util.datetime_to_id(ts)

            yield Info('info', f'Generating folder "{new_id}" under "{parent_id}"...')
            new_meta = {
                'title': export_path[j]['title'],
                'type': 'folder',
                'create': new_id,
                'modify': new_id,
            }
            self.book.meta[new_id] = new_meta
            self.book.toc.setdefault(parent_id, []).append(new_id)
            self.map_id_to_new_id[export_path[j]['id']] = new_id
            parent_id = new_id

        yield from self._insert_to_id(id, parent_id, allow_insert=False)
        return parent_id

    def _insert_to_id(self, id, parent_id, allow_insert=True):
        if id in self.book.toc.get(parent_id, []):
            yield Info('debug', f'Skipped appending "{id}" to "{parent_id}": already in')
            return

        parent = self.book.toc.setdefault(parent_id, [])

        if allow_insert and self.target_index is not None:
            parent.insert(self.target_index, id)
            self.target_index += 1
        else:
            parent.append(id)


def run(root, files, book_id='', *, config=None, no_lock=False, **kwargs):
    start = time.time()

    host = Host(root, config)

    # Fail for invalid book ID
    if book_id not in host.books:
        yield Info('error', f'Invalid book "{book_id}".')
        return

    yield Info('debug', f'Loading book "{book_id}".')

    try:
        book = host.books[book_id]

        if book.no_tree:
            yield Info('error', f'Unable to import to book "{book_id}" ("{book.name}") (no_tree).')
            return

        yield Info('info', f'Impoting to book "{book_id}" ({book.name}).')
        lh = nullcontext() if no_lock else book.get_tree_lock().acquire()
        with lh:
            generator = Importer(book, **kwargs)
            yield from generator.run(files)

    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')
